/*  annoteps.c -- Main program for ANNOTEPS

NAME:
    ANNOTEPS

PURPOSE:
    Main program for ANNOTEPS, to add annotation at arbitrary image
    locations using PostScipt fonts.

EXTERNAL ASSOCIATIONS:
    This program requires the ImageMagick program "convert" to convert a
    PostScript annotation text specification to a raster image.

ARGUMENTS:
    IN		Name of input image (subcommands INPLACE and INOUT).
    OUT		Name of output image (subcommands INOUT and GENIMG).
    ODTYPE	Output image data type (subcommands INOUT and GENIMG).
    NL		Number of lines in output image (subcommand GENIMG).
    NS		Number of samples per output line (subcommand GENIMG).
    IMGBGRND	Image background grey level (subcommand GENIMG).
    TEXT	Annotation text string(s).
    LINEPOS	Line position(s) to insert text string(s).
    SAMPPOS	Sample position(s) to insert text strings(s).
    ALIGN	Interpretation of LINEPOS and SAMPPOS.
    ORIENT	Orientation(s) of text string(s).
    SIZE	Font size(s) of text strint(s).
    TYPEFACE	Typeface family(s) of text string(s).
    STYLE	Style(s) of text string(s).
    WEIGHT	Weight(s) of text strings(s).
    CHARGREY	Grey level(s) of text string(s).
    BACKGRND	Background grey level(s) of text block(s).
    TRANSOPT	Transparent text option(s) for text block(s).

FILE REFERENCES:
    IN			READ to obtain input image (UPDATED for
			subcommand INPLACE).
    OUT			CREATEd to write annotated output image.
    <tmproot>.horiz	CREATEd by aps_bldtxt() to hold rasterized
			horizontal text strings, deleted by
			aps_instxt().
    <tmproot>.vert	CREATEd by aps_bldtxt() to hold rasterized
			vertical text strings, deleted by aps_instxt().

    Additional temporary files of form <tmproot>*.* are created and
    deleted by function aps_bldtxt().

PROGRAM HISTORY:                                  
    2003-12	Initial version, R. A. White, PSU/EMSEI.
    2005-03	Updated for Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.8 and Linux 2.4.

    The program requires the presence of the ImageMagick "convert"
    program -- see also note about APS_CONVERT in annoteps.h.

PROJECT:
    LAS at PSU/EMSEI.

INFORMATIONAL AND ERROR MESSAGES:

    These are described in annoteps.pdf.

EXTERNAL FUNCTIONS CALLED:

(Program Specific Functions)

aps_getarg()	Get program arguments from TAE command interface.

aps_bldtxt()	Build rasterized versions of annotation text strings.

aps_instxt()	Insert rasterized annotation into output image.

    Calling Hierarchy for Program Specific Functions:

	main-annoteps
	    first-level-function-name
		second-level-function-name
		another-second-level-function-name
	    next-first-level-function-name
		(etc)

(Other Non-LAS & Non-ASCI C Functions)

ALGORITHM DESCRIPTION:
    Initialize las/tae interface <c_inlas>
    Get program arguments <aps_getarg>
    Create rasterized blocks containing annotation text <aps_bldtxt>
    Insert annotation into output image <aps_instxt>
    Report processing completion <c_complt>

ALGORITHM REFERENCES:
    PostScript Language Reference, Third Edition, Adobe Systems Inc.
    									
***************   End of Prolog for ANNOTEPS   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O functions		*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"util.h"	/* LAS utility functions		*/
#include	"annoteps.h"	/* Program-specific definitions		*/
									/*
***************   Main Routine for ANNOTEPS   ***************	*/

int main( void )
    {
	                                              			/*
Local Variable Definitions:						*/
    long	ifminmax;	/* Flag set if $MINMAX = "yes" */
    struct APS_IMG
		input;		/* Input image specifications */
    long	nlo;		/* Number of lines in output image */
    long	nso;		/* Number of samples per output line */
    long	odtype;		/* Output data type code */
    float	outbgrnd;	/* Background grey level for generated
				   image */
    char	outimg[CMLEN] = "";
				/* Name of output image file */
    long	nblock;		/* Number of annotation text blocks */
    long	npixhoriz;	/* Width of temporary LTOR text file */
    long	npixvert;	/* Width of temporary vert. text file */
    long	one = 1;	/* Constant one */
    long	pmsgtim;	/* Interval between progress messages */
    long	stdlun;		/* Unit number for standard output */
    long	subcmd;		/* Code for program subcommand */
    long	status;		/* Status returned by function call */
    struct APS_TEXT
		*text;		/* Annotation text block specs */ 
    char	tmproot[CMLEN + 14];
				/* Root filename for temporary files */
    struct                
      PARBLK	vblock;		/* TAE parameter block */

/*  Initialize TAE and LAS  */                                          

    c_inlas (&vblock, &stdlun);

/*  Ingest program arguments  */

    aps_getarg( &vblock, &subcmd, &input, outimg, &odtype, &nlo, &nso,
      &outbgrnd, &text, &nblock, tmproot, &pmsgtim, &ifminmax );

/*  Create rasterized annotation text blocks  */

    if (pmsgtim > 0)
	printf( "ANNOTEPS converting text string(s) to %s\n",
	  "raster text block(s)" );
    aps_bldtxt( nblock, tmproot, nlo, nso, text, &npixhoriz,
      &npixvert ); 

/*  Insert rasterized annotation into output image  */

    status = aps_instxt( &vblock, subcmd, &input, outimg, &odtype, nlo,
      nso, outbgrnd, text, nblock, tmproot, npixhoriz, npixvert,
      pmsgtim, ifminmax );

/*  Write completion message  */

    if (status == E_SUCC)
	{
	if (subcmd == APS_INPLACE)
	    c_complt( "annoteps", (void *) input.hname,
	      &(input.totimage) );
	else
	    c_complt( "annoteps", (void *) outimg, &one );
	}
    else if (subcmd == APS_INPLACE)
	c_lasmsg( TRUE, "annoteps-fatal", "Fatal error encountered" );
    else
	c_clean( &vblock, "annoteps", (void *) outimg, &one );

    return 0;	/* To suppess compiler warning */

    }			/* End of Program ANNOTEPS */
