/*  aps_bldtxt.c -- Create rasterized annotation text strings

NAME:
    aps_bldtxt()

PURPOSE:
    Called by ANNOTEPS to buld PostScript specifications of annotation
    text string(s), invoke ImageMagick "convert" program to create
    rasterized version, determine exact size of each text block and
    compute locations relative to ImageMagick output and LAS image
    window, and, if more than one text block on any image line, assemble
    blocks for faster access during insertion into LAS image.

ARGUMENTS:
    long    nblock	(IN)	Number of annotation text blocks.
    char    *tmproot	(IN)	Root filename for temporary files.
    long    nlo		(IN)	Number of lines in output window.
    long    nso		(IN)	Number of samples in output window.
    struct APS_TEXT
		*text;	(MOD)	Annotation text block specs.
    long    *npixhoriz	(OUT)	Width of temporary LTOR text file.
    long    *npixvert	(OUT)	Width of temporary vert. text file.

RETURN VALUE:
    None.  Function aborts on error.

FILE REFERENCES:
    <tmproot>.ps	CREATEd to write PostScript specification(s) of
			annotation text string(s), where <tmproot> is
			tmp.las_$SESSION, in the output directory (for
			subcommands INOUT and GENIMG) or input directory
			(subcommand INPLACE); deleted before function
			returns.
    <tmproot>.gray	CREATEd by ImageMagick "convert" program to
			contain annotation text block(s); deleted before
			function returns.
    <tmproot>.horiz	CREATEd to hold rasterized horizontal text
			strings.
    <tmproot>.vert	CREATEd to hold rasterized vertical text
			strings.

PROGRAM HISTORY:                                  
    2003-11	Initial version, R. A. White, PSU/EMSEI.
    2005-03	Updated for Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.8 and Linux 2.4.

    The program requires the presence of the ImageMagick "convert"
    program -- see also note about APS_CONVERT in annoteps.h.

PROJECT:
    LAS at PSU/EMSEI.

INFORMATIONAL AND ERROR MESSAGES:

    These are described in annoteps.pdf.

EXTERNAL FUNCTIONS CALLED:

    ImageMagick "convert" program to convert PostScript specifications
    of text strings to raster representation.

ALGORITHM DESCRIPTION:
    Build separate tables of horizontal and vertical text strings
    For horizontal and then vertical strings
	Setup starting row and column for each string
	Get maximum length of any string
	Open PostScript specification file
	Write header information, including bounding box
	For each text string
	    Write font specification if not done for previous string
	    Write PostScript starting point for string
	    Write rotation if vertical string
	    Write text
	    Restore rotational geometry
	Close PostScript specification file
	Invoke ImageMagick convert to create raster
	Delete PostScript specification file
	Open rasterized text file
	For each string
	    Find edges of bounding box
	    Pad bounding box if OPAQUE option specified
	    Compute edges of box relative to LAS image window
	    Check for box within image window
	Close rasterized text file
	Change extension of rasterized text file to .horiz or .vert
	Save width of rasterized text file

ALGORITHM REFERENCES:
    PostScript Language Reference, Third Edition, Adobe Systems Inc.
    									
***************   End of Prolog for aps_bldtxt()   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O functions		*/
#include	<string.h>	/* String manipulation functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"annoteps.h"	/* Program-specific definitions		*/

									/*
***************   Start of Function aps_bldtxt()   ***************	*/

FUNCTION void aps_bldtxt(
    long	nblock,		/* Number of annotation text blocks */
    char	*tmproot,	/* Root filename for temporary files */
    long	nlo,		/* Number of lines in output window */
    long	nso,		/* Number of samples in output window */
    struct APS_TEXT
		*text,		/* Annotation text block specs */
    long	*npixhoriz,	/* Width of temporary LTOR text file */
    long	*npixvert)	/* Width of temporary vert. text file */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	basepos;	/* Position of base of text string */
    long	block;		/* Index of current text block */
    long	blockindex[2][APS_MAXSTRING];
				/* Index number(s) of horizontal and
				   vertical text string(s) */
    long	bounds[APS_MAXSTRING][4];
				/* Top, bottom, left, and right
				   boundaries of each rasterized text
				   string, counting from 1 */ 
    long	col;		/* Column of rasterized text file */
    char	command[300];	/* Buffer for system() command */
    char	*facenames[] = APS_FACENAMETABLE;
				/* Array of supported typeface names */
    long	flag;		/* Flag set if text-block pixel found */
    long	fontno;		/* Number of PS defined font */
    long	height;		/* Height of PostScript bounding box */
    char	hfname[CMLEN + 20];
				/* Name of horizontal string file */
    long	i;		/* Loop counter/index */
    long	index;		/* Index into psnames table */
    long	line;		/* Line of raster text file or image */
    long	orient;		/* Current string orientation: 0 for
				   horizontal, 1 for vertical */
    long	orientcnt[2];	/* Number of horizontal and vertical
				   text strings */
    long	pad;		/* Padding around rasterized text */ 
    FILE	*psfile;	/* File pointer for PostScript file */
    char	psfname[132];	/* Name of temporary PostScript file */
    char	*psnames[] = APS_PSNAMETABLE;
				/* Array of PostScript-recognized names
				   for each typeface in regular, oblique,
				   bold, and bold-oblique versions */
    long	psnindex[APS_MAXSTRING];
				/* Indices into psnames table */
    char	*rastbuf;	/* Buffer for rasterized text */
    FILE	*rifile;	/* File pointer for raster text file */
    char	rifname[132];	/* Name of raster text file */
    long	sizetable[APS_MAXSTRING];
				/* Sizes of fonts in psnindex table */
    long	status;		/* Status returned by system() call */
    char	vfname[CMLEN + 20];
				/* Name of vertical string file */
    long	width;		/* Width of PostScript bounding box */

/*  Build separate tables of horizontal and vertical text strings  */

    for (i = orientcnt[0] = orientcnt[1] = 0; i < nblock; i++)
	{
	if (text[i].text[0] == '\0')
	    {
	    c_lasmsg( FALSE, "annoteps-nullstr", "Text string %d%s",
	      i + 1, "is all blanks -- ignoring it" );
	    continue;
	    }
	if (text[i].orient == APS_LTOR)
	    blockindex[0][orientcnt[0]++] = i;
	else
	    blockindex[1][orientcnt[1]++] = i;
	}

/*  Process horizontal strings and then vertical strings  */

    for (orient = 0; orient < 2; orient++)
	{
	if (orientcnt[orient] < 1)
	    continue;
	psfile = rifile = NULL;

/*  Setup starting row and column for each string, and get bounds
    Set block to extend 80% of size above baseline, and 40% below  */

	height = 10;
	width = 10;
	for (i = 0; i < orientcnt[orient]; i++)
	    {
	    block = blockindex[orient][i];
	    if (orient == APS_LTOR)		/* horizontal text */
		{
		text[block].nl = 1.2 * text[block].size + 2;
		text[block].topl = height - 5;
		height += text[block].nl;
		text[block].ns
		  = text[block].size * strlen( text[block].text );
		if (text[block].ns + 10 > width)
		    width = text[block].ns + 10;
		}
	    else			/* vertical text */
		{
		text[block].ns = 1.2 * text[block].size + 2;
		text[block].lefts = width - 5;
		width += text[block].ns;
		text[block].nl
		  = text[block].size * strlen( text[block].text );
		if (text[block].nl + 10 > height)
		    height = text[block].nl + 10;
		}
	    }

/*  Open PostScript specification file  */

	sprintf( psfname, "%s.ps", tmproot );
	if ((psfile = fopen( psfname, "w" )) == NULL)
	    c_lasmsg( TRUE, "annoteps-tmpopen",
	      "Error opening temporary file %s", psfname );

/*  Write header information, including bounding box  */

	fprintf( psfile, "%s\n", "%!PS-Adobe-3.0" );
	fprintf( psfile, "%s %d %d %ld %ld\n", "%%BoundingBox:", 0, 0,
	  width, height );
	fprintf( psfile, "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n\n",
	  "%%DocumentData: Clean7Bit", "%%Orientation: Portrait",
	  "%%Pages: 1", "%%PageOrder: Ascend", "%%Title: ",
	  "%%EndComments", "%%BeginProlog",
	  "%%BeginResource: procset Enscript-Encoding-88591 1.6 1" );
	fprintf( psfile, "%s\n%s\n\n",
	  "/FSD {findfont exch scalefont def} bind def",
	  "/SMS {setfont moveto show} bind def" );

/*  Write PostScript specification of each text string  */

	for (i = 0; i < APS_MAXSTRING; i++)
	    psnindex[i] = -1;
	for (i = 0; i < orientcnt[orient]; i++)
	    {
	    block = blockindex[orient][i];

/*  Write font specification if not yet done  */

	    for (fontno = 0; fontno < APS_FACECOUNT; fontno++)
		if (strcmp( text[block].typeface, facenames[fontno] )
		  == 0)
		    break;
	    index= 4 * fontno + text[block].style
	      + 2 * text[block].weight;
	    for (fontno = 0; psnindex[fontno] >= 0; fontno++)
		if ((psnindex[fontno] == index)
		  && (sizetable[fontno] == text[block].size))
		    break;
	    if (psnindex[fontno] < 0)
		{
		psnindex[fontno] = index;
		sizetable[fontno] = text[block].size;
		fprintf( psfile, "/F%ld %ld /%s FSD\n", fontno,
		  sizetable[fontno], psnames[index]);
		}

/*  Specify PostScript coordinates of start of text block  */

	    if (text[block].orient == APS_LTOR)
		{
		text[block].lefts = 5;
		basepos =  height
		  - (text[block].topl + 0.8 * text[block].size);
		fprintf( psfile, "%d %ld moveto\n", 5, basepos ); 
		}
	    else		/* vertical text */
		{
		text[block].topl = 5;
		if (text[block].orient == APS_TTOB)
		    {
		    basepos
		      = text[block].lefts + (text[block].ns - 1)
		       - 0.8 * text[block].size; 
		    fprintf( psfile, "%ld %ld moveto\n", basepos,
		      height - 5 );
		    fprintf( psfile, "270 rotate\n" );
		    }
		else		/* text reads bottom to top */
		    {
		    basepos
		      = text[block].lefts + 0.8 * text[block].size; 
		    fprintf( psfile, "%ld %ld moveto\n", basepos,
		      height - 5 - text[block].nl );
		    fprintf( psfile, "90 rotate\n" );
		    }
		}

/*  Specify text to be created  */

	    fprintf( psfile, "F%ld setfont\n", fontno );
	    fprintf( psfile, "(%s) show\n", text[block].text );
	    if (orient != APS_LTOR)
		fprintf( psfile, "grestore\n\n" );
	    else
		fprintf( psfile, "\n" );
	    }

/*  Close PostScript specification file  */

	fclose( psfile );

/*  Invoke ImageMagick "convert" to create rasterized text  */

	sprintf( rifname, "%s.gray", tmproot );
	sprintf( command, "%s %s %s", APS_CONVERT, psfname,
	  rifname ); 
	if ((status = system( command )) != 0)
	    c_lasmsg( TRUE, "annoteps-conv", "Error converting %s%s",
	      psfname, " to raster" );

/*  Open rasterized text file after deleting PS file  */
	
	remove( psfname );  
	if ((rifile = fopen( rifname, "r" )) == NULL)
	    c_lasmsg( TRUE, "annoteps-tmpopen",
	      "Error opening temporary file %s", rifname );
	if ((rastbuf = (char*) malloc( width )) == NULL)
	    c_lasmsg( TRUE, "annoteps-malloc",
	      "Memory allocation error" ); 

/*  Find bounding box for each string  */

	for (i = 0; i < orientcnt[orient]; i++)
	    {
	    bounds[i][0] = height;
	    bounds[i][1] = bounds[i][3] = 0;
	    bounds[i][2] = width;
	    }
	for (line = 1; line <= height; line++)
	    {
	    if (fread( rastbuf, 1, width, rifile ) != width)
		c_lasmsg( TRUE, "annoteps-readtmp", "Error reading from"
		  " temporary file %s", psfname );
	    for (i = 0; i < orientcnt[orient]; i++)
		{
		block = blockindex[orient][i];
		flag = 0;
		if ((text[block].topl > line)
		  || (text[block].topl + text[block].nl <= line))
		    continue;
		for (col = text[block].lefts;
		  col < text[block].lefts + text[block].ns; col++)
		    {
		    if (rastbuf[col - 1] == '\377')
			continue;
		    flag = 1;
		    if (col < bounds[i][2])
			bounds[i][2] = col;
		    else if (col > bounds[i][3])
			bounds[i][3] = col;
		    }
		if (flag == 0)
		    continue;
		if (line < bounds[i][0])
		    bounds[i][0] = line;
		else if (line > bounds[i][1])
		    bounds[i][1] = line;
		}
	    }

/*  Pad edges of bounding boxes and enter into text structure  */

	for (i = 0; i < orientcnt[orient]; i++)
	    {
	    block = blockindex[orient][i];
	    if (text[block].transopt != APS_TRANS)
		{
		pad = (orient == APS_LTOR) ?
		  (bounds[i][1] - bounds[i][0])
		  : (bounds[i][3] - bounds[i][2]);
		pad = 1.5 + 0.05 * pad;
		bounds[i][0] -= pad;
		bounds[i][1] += pad;
		bounds[i][2] -= pad;
		bounds[i][3] += pad;
		}
	    text[block].nl = 1 + bounds[i][1] - bounds[i][0];
	    text[block].ns = 1 + bounds[i][3] - bounds[i][2];
	    text[block].topl = bounds[i][0];
	    text[block].lefts = bounds[i][2];

/*  Compute edges of box relative to LAS image window  */

	    if (text[block].orient == APS_LTOR)
		{
		text[block].slo = text[block].linepos
		  - (text[block].nl - 1)
		   * 0.5 * (text[block].align / 3);
		text[block].sso = text[block].samppos
		  - (text[block].ns - 1)
		   * 0.5 * (text[block].align % 3); 
		}
	    else if (text[block].orient == APS_TTOB)
		{
		text[block].slo = text[block].linepos
		  - (text[block].nl - 1)
		   * 0.5 * (text[block].align % 3);
		text[block].sso = text[block].samppos
		  - (text[block].ns - 1)
		   * 0.5 * (2 - text[block].align / 3); 
		}
	    else if (text[block].orient == APS_BTOT)
		{
		text[block].slo = text[block].linepos
		  - (text[block].nl - 1)
		   * 0.5 * (2 - text[block].align % 3);
		text[block].sso = text[block].samppos
		  - (text[block].ns - 1)
		   * 0.5 * (text[block].align / 3); 
		}
	    if ((text[block].slo <= 0) || (text[block].sso <= 0)
	      || ((text[block].slo + text[block].nl - 1) > nlo)
	      || ((text[block].sso + text[block].ns - 1) > nso))
		{
		fclose( rifile );
		remove( rifname );
		if ((orient > 0) && (orientcnt[0] > 0))
		    remove( hfname );
		c_lasmsg( TRUE, "annoteps-outside", "Text string %d"
		  " extends outside image window:  upper left corner"
		  " (line, sample) = (%d, %d); lower right corner = (%d,"
		  " %d)", block + 1, text[block].slo, text[block].sso,
		  text[block].slo + text[block].nl -1,
		  text[block].sso + text[block].ns -1 );
		}
	    }

/*  Close files and unsqueezed version  */

	fclose( rifile );
	free( rastbuf );

/*  Change raster file name extension from .gray and save width  */

	if (orient == 0)
	    {
	    *npixhoriz = width;
	    sprintf( hfname, "%s.horiz", tmproot );
	    if (rename( rifname, hfname ) != 0)
		c_lasmsg( TRUE, "annoteps-tmpren",
		  "Error renaming temporary file %s to %s", rifname,
		  hfname );
	    }
	else
	    {
	    *npixvert = width;
	    sprintf( vfname, "%s.vert", tmproot );
	    if (rename( rifname, vfname ) != 0)
		c_lasmsg( TRUE, "annoteps-tmpren",
		  "Error renaming temporary file %s to %s", rifname,
		  vfname );
	    }
	}

    }			/* End of Function aps_bldtxt() */
