/*  aps_getarg.c -- Get arguments for ANNOTEPS

NAME:
    aps_getarg.c

PURPOSE:
    Called by main program annoteps.c to get program arguments from TAE
    interface.

ARGUMENTS:
    struct PARBLK
	    *vblock	(IN)	TAE parameter block.
    long    *subcmd	(OUT)	Code for program subcommand.
    struct APS_IMG
	    *input	(OUT)	Input image specifications.
    char    *outimg	(OUT)	Name of output image file.
    long    *odtype	(OUT)	Output data type code.
    long    *nlo	(OUT)	Number of lines in output image.
    long    *nso	(OUT)	Number of samples per output line.
    float   *outbgrnd	(OUT)	Background grey level for generated
				image.
    struct APS_TEXT
	    **text      (OUT)	Annotation text block specifications.
    long    *nblock	(OUT)	Number of annotation text blocks.	    
    char    *tmproot	(OUT)	Root filename for temporary files.
    long    *pmsgtim	(OUT)	Interval between progress messages.
    long    *ifminmax	(OUT)	Flag set if $MINMAX = "yes".

RETURN VALUE:
    None.  Function aborts on error.

PROGRAM HISTORY:                                  
    2003-10	Initial version, R. A. White, PSU/EMSEI.
    2005-03	Updated for Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.8 and Linux 2.4.

PROJECT:
    LAS at PSU/EMSEI.

INFORMATIONAL AND ERROR MESSAGES:

    These are described in annoteps.pdf.

ALGORITHM DESCRIPTION:
    Get subcommand from TAE argument block.
    If subcommand is INPLACE or INOUT
	Allocate temporary space for input specifications
	Get name(s) and specification(s) of logical input image
	Allocate space in input-image structure
	Copy input specifications to input-image structure
	Free temporary space
    If subcommand is INOUT or GENIMG
	Get name and data type of output image
    If subcommand is GENIMG
	Get size and background grey level for output image
    Create root for temporary files
    Allocate space for specifications of annotation text strings
    For each text string
	Save string
	Get postion and orientation
	Get font specifications
	Get grey-levels and build look-up table
	Get transparency option
    Get interval between processing progress messages
    Get min/max option
    									
***************   End of Prolog for aps_getarg.c   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O functions		*/
#include	<string.h>	/* String manipulation functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"util.h"	/* LAS utility functions		*/
#include	"annoteps.h"	/* Program-specific definitions		*/
									/*
External Function Prototypes Not Included in Header Files:		*/
    struct VARIABLE *p_find( struct PARBLK *vblock, char *vname );
									/*
***************   Start of Function aps_getarg   ***************	*/

FUNCTION void aps_getarg(
    struct PARBLK
		*vblock,	/* TAE parameter block */
    long	*subcmd,	/* Code for program subcommand */
    struct APS_IMG
		*input,		/* Input image specifications */
    char	*outimg,	/* Name of output image file */
    long	*odtype,	/* Output data type code */
    long	*nlo,		/* Number of lines in output image */
    long	*nso,		/* Number of samples per output line */
    float	*outbgrnd,	/* Background grey level for generated
				   image */
    struct APS_TEXT
		**text,		/* Annotation text block specs */
    long	*nblock,	/* Number of annotation text blocks */
    char	*tmproot,	/* Root filename for temporary files */
    long	*pmsgtim,	/* Progress message interval */
    long	*ifminmax)	/* Flag set if $MINMAX = "yes" */
    {
	                                              			/*
Local Variable Definitions:						*/

    struct VARIABLE
		*argptr;	/* Pointer to argument in vblock */
    char	*cptr;		/* Pointer to character in string */
    char	dtypename[5];	/* Name of output data type */
    long	i;		/* Loop counter/index; temp variable */
    long	imglimit;	/* Maximum images in logical image */
    long	j;		/* Loop counter/index */
    long	maxlen;		/* Maximum length of text string */
    long	nimg;		/* Number of images in logical image */
    long	nval;		/* Number of argument values */
    char	odoption[]	/* Possible values of ODTYPE argument */
		  = "SAMEBYTEI*2 I*4 R*4";
    long	one = 1;	/* Constant one */
    long	qflag[MAXIMG][2];/* Flags for window defaults */
    long	same_flag;	/* Flag for input windows all same */
    float	step;		/* Step in grey-level look-up table */
    long	*tmpbnd;	/* Pointer into tmpbuf for bands */
    char	*tmpbuf;	/* Temporary buffer for image args */
    char	*tmphnam;	/* Pointer in tmpbuf for host names */
    long	*tmpnbnd;	/* Pointer into tmpbuf for nbands */
    long	*tmpwin;	/* Pointer into tmpbuf for window */
    char	*tmptnam;	/* Pointer into tmpbuf for TAE names */
    long	zero = 0;	/* Constant zero */

/*  Get subcommand from TAE argument block  */

    argptr = p_find( vblock, "_SUBCMD" );
    if (strcmp( "INPLACE", SVAL( *argptr, 0 ) ) == 0)
	*subcmd = APS_INPLACE;
    else if (strcmp( "INOUT", SVAL( *argptr, 0 ) ) == 0)
	*subcmd = APS_INOUT;
    else
	*subcmd = APS_GENIMG;

/*  Get specifications of input image  */

    if ((*subcmd == APS_INPLACE) || (*subcmd == APS_INOUT))
	{
	
/*  First allocate temporary buffer space for image specifications  */

	if ((tmpbuf = (char *) malloc( 2 * MAXIMG * CMLEN
	  + 4 * MAXIMG * sizeof( long )
	  + MAXIMG * (MAXBND + 1) * sizeof( long )
	  + MAXIMG * sizeof( long ) )) == NULL)
	    c_lasmsg( TRUE, "annoteps-malloc",
	      "Memory allocation error" );
	tmpwin = (long *) tmpbuf;
	tmpbnd = tmpwin + 4 * MAXIMG;
	tmpnbnd = tmpbnd + MAXIMG * (MAXBND + 1);
	tmphnam = (char *) (tmpnbnd + MAXIMG);
	tmptnam = tmphnam + MAXIMG * CMLEN;

/*  Get name(s) and size(s) of input image(s)  */

	imglimit = MAXIMG;
	if (c_getin( vblock, "IN", &zero, (void *) tmptnam,
	  (void *) tmphnam, &(input->totimage), (void *) tmpwin,
	  (void *) tmpbnd, tmpnbnd, &(input->totband), &imglimit,
	  &same_flag, qflag )
	   != E_SUCC)
	    c_lasmsg( TRUE, "annoteps-fatal", "" );
	if (same_flag == 0) c_lasmsg( TRUE, "annoteps-diffwin",
	      "Mismatched input windows" );
	if (*subcmd == APS_INPLACE)
	    {
	    for (i = 0; i < input->totimage; i++)
		if (qflag[i][0])
		    c_lasmsg( TRUE, "annoteps-inplwin",
		      "Windowing not allowed for subcommand INPLACE" );
	    }

/*  Allocate array space in APS_IMG structure for image  */

	nimg = input->totimage;
	if ( ((input->hname
	   = (char *) malloc( nimg * CMLEN )) == NULL)
	  || ((input->dtype
	   = (long *) malloc( nimg * sizeof( long ) )) == NULL)
	  || ((input->dlength
	   = (long *) malloc( nimg * sizeof( long ) )) == NULL)
	  || ((input->window
	   = (long *) malloc( nimg * 4 * sizeof( long ) )) == NULL)
	  || ((input->nband
	   = (long *) malloc( nimg * sizeof( long ) )) == NULL)
	  || ((input->bands
	   = (long *) malloc( nimg * (MAXBND + 1) * sizeof( long * ) ))
	   == NULL)
	  || ((input->offsets
	   = (long *) malloc( input->totband * sizeof( long ) ))
	   == NULL) ) 
	    c_lasmsg( TRUE, "annoteps-malloc",
	      "Memory allocation error" );

/*  Move arrays from temporary buffer to APS_IMG structure for image  */

	for (j = 0; j < nimg; j++)
	    {
	    strcpy( input->hname + j * CMLEN, tmphnam + j * CMLEN );
	    memcpy( input->window + 4 * j, tmpwin + 4 * j,
	      4 * sizeof( long ) );
	    input->nband[j] = tmpnbnd[j];
	    memcpy( input->bands + j * (MAXBND + 1),
	      tmpbnd + j * (MAXBND + 1), tmpnbnd[j] * sizeof( long ) );
	    *(input->bands + j * (MAXBND + 1) + input->nband[j]) = 0;
	    }
	*nlo = input->window[2];
	*nso = input->window[3];
	}
    else if ((tmpbuf = (char *) malloc( CMLEN )) == NULL)
	c_lasmsg( TRUE, "annoteps-malloc", "Memory allocation error" );


/*  Get output image specifications  */		 

    if ((*subcmd == APS_INOUT) || (*subcmd == APS_GENIMG))
	{
	if (c_getout( vblock, "OUT", (void *) tmpbuf, (void *) outimg )
	  != E_SUCC)
	    c_lasmsg( TRUE, "annoteps-fatal", "" );
	i = 5;
	if (c_getstr( vblock, "ODTYPE", &one, &i, dtypename ) != one)
	    c_lasmsg( TRUE, "annoteps-fatal", "" );
	*odtype = (strstr( odoption, dtypename ) - odoption ) / 4;
	}
    if (*subcmd == APS_GENIMG)
	{
	*outbgrnd = RVAL( *p_find (vblock, "OUTBGRND" ), 0 );
	*nlo = IVAL( *p_find (vblock, "NL" ), 0 );
	*nso = IVAL( *p_find (vblock, "NS" ), 0 );
	}

/*  Create root for temporary files  */

    if (*subcmd == APS_INPLACE)
	strcpy( tmproot, input->hname );
    else
	strcpy( tmproot, outimg );
    if ((cptr = strrchr( tmproot, (int) '/' ))== NULL)
	cptr = tmproot;
    *cptr = '\000';
    strcat( tmproot, "tmp.las_" );
    strcat( tmproot, SVAL(  *p_find( vblock, "$SESSION" ), 0 ) );

/*  Get specifications of annotation text strings  */

/*  First allocate space for descriptors for each string  */

    argptr = p_find( vblock, "TEXT" );
    *nblock = (*argptr).v_count;
    if ((*text = (struct APS_TEXT *)
      malloc( *nblock * sizeof( struct APS_TEXT ) )) == NULL)
        c_lasmsg( TRUE, "annoteps-malloc", "Memory allocation error" );

/*  Get text of each annotation string  */

    maxlen = 132;
    for (i = 0; i < *nblock; i++)
	strcpy( (*text)[i].text, SVAL( *argptr, i ) );

/*  Get position and orientation of each string  */

    argptr = p_find( vblock, "LINEPOS" );
    if ((nval = (*argptr).v_count) < *nblock)
	c_lasmsg( TRUE, "annoteps-fewpos",
	  "Too few values for LINEPOS or SAMPPOS" );
    for (i = 0; i < *nblock; i++)
	(*text)[i].linepos = IVAL( *argptr, i );
    argptr = p_find( vblock, "SAMPPOS" );
    if ((nval = (*argptr).v_count) < *nblock)
	c_lasmsg( TRUE, "annoteps-fewpos",
	  "Too few values for LINEPOS or SAMPPOS" );
    for (i = 0; i < *nblock; i++)
	(*text)[i].samppos = IVAL( *argptr, i );
    argptr = p_find( vblock, "ALIGN" );
    nval = (*argptr).v_count;
    strcpy( tmpbuf, "TL TC TR CL CC CR BL BC BR" );
    for (i = 0; i < *nblock; i++)
	(*text)[i].align = (i < nval) ?
	  (strstr( tmpbuf, SVAL( *argptr, i ) ) - tmpbuf) / 3 :
	  (*text)[i - 1].align;
    argptr = p_find( vblock, "ORIENT" );
    nval = (*argptr).v_count;
    strcpy( tmpbuf, "LTOR TTOB BTOT" );
    for (i = 0; i < *nblock; i++)
	(*text)[i].orient = (i < nval) ?
	  (strstr( tmpbuf, SVAL( *argptr, i ) ) - tmpbuf) / 5 :
	  (*text)[i - 1].orient;

/*  Get font specifications for each string  */

    argptr = p_find( vblock, "SIZE" );
    nval = (*argptr).v_count;
    for (i = 0; i < *nblock; i++)
	(*text)[i].size = (i < nval) ? IVAL( *argptr, i )
	  : (*text)[i - 1].size;
    argptr = p_find( vblock, "TYPEFACE" );
    nval = (*argptr).v_count;
    for (i = 0; i < *nblock; i++)
	strcpy( (*text)[i].typeface, (i < nval) ? SVAL( *argptr, i ) :
	  (*text)[i - 1].typeface );
    argptr = p_find( vblock, "STYLE" );
    nval = (*argptr).v_count;
    strcpy( tmpbuf, "REGULAR OBLIQUE" );
    for (i = 0; i < *nblock; i++)
	(*text)[i].style = (i < nval) ?
	  (strstr( tmpbuf, SVAL( *argptr, i ) ) - tmpbuf) / 8 :
      (*text)[i - 1].style;
    argptr = p_find( vblock, "WEIGHT" );
    nval = (*argptr).v_count;
    strcpy( tmpbuf, "REGULAR BOLD" );
    for (i = 0; i < *nblock; i++)
	(*text)[i].weight = (i < nval) ?
	  (strstr( tmpbuf, SVAL( *argptr, i ) ) - tmpbuf) / 8 :
	  (*text)[i - 1].weight;

/*  Get grey-level range and build look-up table for each string  */

    argptr = p_find( vblock, "CHARGREY" );
    nval = (*argptr).v_count;
    for (i = 0; i < *nblock; i++)
	(*text)[i].chargrey = (i < nval) ? RVAL( *argptr, i )
	  : (*text)[i - 1].chargrey;
    argptr = p_find( vblock, "BACKGRND" );
    nval = (*argptr).v_count;
    for (i = 0; i < *nblock; i++)
	{
	(*text)[i].backgrnd = (i < nval) ? RVAL( *argptr, i )
	  : (*text)[i - 1].backgrnd;
	step = ((*text)[i].chargrey - (*text)[i].backgrnd) / 255;
	for (j = 0; j < 256; j++)	/* PS background = 255 */
	    (*text)[i].lookup[255 - j] = (*text)[i].backgrnd + j * step;
	}

/*  Get transparency option for each string  */

    argptr = p_find( vblock, "TRANSOPT" );
    nval = (*argptr).v_count;
    strcpy( tmpbuf, "TRANSPARENT OPAQUE" );
    for (i = 0; i < *nblock; i++)
	(*text)[i].transopt = (i < nval) ?
	  (strstr( tmpbuf, SVAL( *argptr, i ) ) - tmpbuf) / 12 :
      (*text)[i - 1].transopt;

/*  Get global variables  */

    *pmsgtim = IVAL( *p_find( vblock, "$MSGTIM" ), 0);
    *ifminmax
      = (strcmp( SVAL( *p_find( vblock, "$MINMAX" ), 0 ), "YES" ) == 0)
       ? 1 : 0;
    if ((*subcmd == APS_INPLACE) && (*ifminmax))
	{
	*ifminmax = 0;
	c_lasmsg( FALSE, "", "Not computing new minimum/maximum for"
	  " image -- use LAS program MINMAX if min/max needed" );
	}
    free( tmpbuf );

    return;

    }			/* End of Function aps_getarg */
