/*  aps_instxt.c -- Insert text strings into image

NAME:
    aps_instxt()

PURPOSE:
    Called by ANNOTEPS to insert rasterized text strings into the
    image.

ARGUMENTS:
    struct PARBLK
	    *vblock	(IN)	TAE parameter block.
    long    subcmd	(IN)	Code for program subcommand.
    struct APS_IMG
	    *input	(MOD)	Input image specifications.
    char    *outimg	(IN)	Name of output image file.
    long    *odtype	(MOD)	Output data type code.
    long    nlo		(IN)	Number of lines in output image.
    long    nso		(IN)	Number of samples per output line.
    float   outbgrnd	(IN)	Background grey level for generated
				   image.
    struct APS_TEXT
	    *text	(IN)	Annotation text block specs. 
    long    nblock	(IN)	Number of annotation text blocks.
    char    *tmproot	(IN)	Root filename for temporary files.
    long    npixhoriz	(IN)	Width of temporary LTOR text file.
    long    npixvert	(IN)	Width of temporary vert. text file.
    long    pmsgtim	(IN)	Interval between progress messages.
    long    ifminmax	(IN)	Flag set if $MINMAX = "yes".

RETURN VALUE:
    E_SUCC	Successful completion.
    E_FAIL	Fatal error occurred.

FILE REFERENCES:
    <input>->hname[]	READ for subcommand INOUT,
			UPDATED with text strings for subcommand
			INPLACE.
    <outimg>		CREATED for subcommand INOUT and GENIMG.
    <tmproot>.horiz	READ and DELETEd to process rasterized
			horizontal text strings.
    <tmproot>.vert	READ and DELETEd to process rasterized
			vertical text strings.

PROGRAM HISTORY:                                  
    2003-12	Initial version, R. A. White, PSU/EMSEI.
    2005-03	Updated for Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.8 and Linux 2.4.

PROJECT:
    LAS at PSU/EMSEI.

INFORMATIONAL AND ERROR MESSAGES:

    These are described in annoteps.pdf.

ALGORITHM DESCRIPTION:

    Set up names of rasterized text files
    Open rasterized text files
    Allocate buffer for rasterized text
    If subcommand is INPLACE or INOUT
	Open input image(s)
    If subcommand is INOUT or GENIMG
	Open output image
    Allocate image buffers
    Set up image I/O group
    If minimum/maximum calculation wanted
	Initialize extrema for each output band
    If processing progress message wanted
	Initialize processing message
	Start processing messages
    Read first input image line, if any, and initialize output buffer
    For each image line
	If subcommand is INOUT
	    Copy input line(s) to output buffer
	Else if subcommand is GENIMG and background needs to be set
	    Create output line filled with IMGBGRND value
	For each text string
	    If current image line within block containing string
		Set flag specifying image line modified
		Position rasterized text file to start of text line
		Read line from text block into buffer
		For each image band
		    Set pointer to start of text in output image
		    For each pixel in text line
			If opaque mode or pixel is not text background
			    Convert pixel value using lookup table
			    Insert value into image line
	Update maxima and minima for each output band
	Write output line and read next input, if any
    If processing progress messages
        Stop them
    Close all files
    If subcommand is INOUT
	Copy history information from input
	Update DDR
    If min/max compiled
	Enter values into DDR
    									
***************   End of Prolog for APS_INSTXT   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O functions		*/
#include	<string.h>	/* String manipulation functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"pixman.h"	/* Pixel manipulation functions		*/
#include	"typlim.h"	/* LAS value limits for data types	*/
#include	"util.h"	/* LAS utility functions		*/
#include	"annoteps.h"	/* Program-specific definitions		*/
						   			/*
Local Program Constants and Macros:					*/
/*	if (textptr[orient] != NULL) \
	fclose( textptr[orient] ); \*/

#define APS_REMTEXT		/* Remove rasterized text files  */ \
    for (orient = 0; orient < 2; orient++) \
	{ \
	if ((textwidth[orient] == 0) || (textptr[orient] == NULL)) \
	    continue; \
	remove( textfname[orient] ); \
	} \
    if (gdesc != NULL) \
	c_egclse( &gdesc )
									/*
External Variable Declarations:						*/
    extern	long	curent;	/* Current entity for processing
				   progress meassages */
    extern	long	curcat;	/* Current category for processing
				   progress meassages */
									/*
***************   Start of Function aps_instxt()   ***************	*/

FUNCTION long aps_instxt(
    struct PARBLK
		*vblock,	/* TAE parameter block */
    long	subcmd,		/* Code for program subcommand */
    struct APS_IMG
		*input,		/* Input image specifications */
    char	*outimg,	/* Name of output image file */
    long	*odtype,	/* Output data type code */
    long	nlo,		/* Number of lines in output image */
    long	nso,		/* Number of samples per output line */
    float	outbgrnd,	/* Background grey level for generated
				   image */
    struct APS_TEXT
		*text,		/* Annotation text block specs */ 
    long	nblock,		/* Number of annotation text blocks */
    char	*tmproot,	/* Root filename for temporary files */
    long	npixhoriz,	/* Width of temporary LTOR text file */
    long	npixvert,	/* Width of temporary vert. text file */
    long	pmsgtim,	/* Interval between progress messages */
    long	ifminmax)	/* Flag set if $MINMAX = "yes" */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	acc;		/* Code for type of image access */
    long	band;		/* Current band in total bands */
    long	bginitflag;	/* Flag set to when output line needs to
				   be initialized with background */ 
    long	block;		/* Current rasterized text block */
    double      dblone = 1.0;   /* Constant value 1 */
    long	dlengths[9] = {0, EBYTESIZE, EWORDSIZE, ELONGSIZE,
		  EREALSIZE, EDOUBLESIZE, ECHARSIZE, EUWORDSIZE,
		  EULONGSIZE};	/* Lengths of data types */
    double	dmaxval[5] = {0, MAXUTINY, MAXWORD, MAXLONG, MAXREAL};
				/* Maximum values of data types */
    double	dminval[5] = {0, MINUTINY, MINWORD, MINLONG, MINREAL};
				/* Minimum values of data types */
    char	*dnames[] = {"same", "byte", "integer*2", "integer*4",
		  "real*4"};	/* Names of data types for message */
    long	dtype;		/* Data type for current image buffer */
    struct FDESC*
		fdesc[MAXIMG+2];/* Image file I/O descriptors */
    struct GDESC*
		gdesc = NULL;	/* Image I/O group descriptors */
    long	iband;		/* Current band within input image */
    long	imgfilecnt;	/* Total number of image files */
    long	inimage;	/* Number of current input image */
    unsigned char
		*imgbuf;	/* Buffer for image I/O */
    long	imgbufsiz;	/* Size of buffer for image I/O */
    long	line;		/* Current output line */
    double	maxval[MAXBND]; /* Maximum value for each band */
    double	minval[MAXBND]; /* Minimum value for each band */
    long	nbandout;	/* Number of output bands */
    long	odlength = 0;	/* Length of output pixels, in bytes */
    long	one = 1;	/* Constant one */
    long	opt[2] = {0,0};	/* I/O option flags */
    long	orient;		/* Text orientation:  horiz. or vert. */
    long	outoffs[MAXBND];/* Array of output buffer offsets */
    long	pixoverflag[2];	/* Flags for text pixel und/overflow */
    unsigned char
		*outptr;	/* Pointer into output buffer */
    unsigned char
		*textbuf;	/* Buffer for rasterized text */
    char	textfname[2][CMLEN + 20];
				/* Names of rasterized text files */
    FILE	*textptr[2];	/* File pointers for rasterized text */
    double	textval;	/* Value of text pixel after table
				   lookup */
    long	textwidth[2];	/* Width of text files */
    unsigned char
		*tpixel;	/* Pointer to next pixel in text line */
    long	zero = 0;	/* Constant zero */

/*  Open rasterized text file(s)  */

    sprintf( textfname[0], "%s.horiz", tmproot );
    sprintf( textfname[1], "%s.vert", tmproot );
    for (orient = 0; orient < 2; orient++)
        {
        textwidth[orient] = (orient == 0) ? npixhoriz : npixvert;
        if (textwidth[orient] > 0)
            if ((textptr[orient] = fopen( textfname[orient], "r" ))
	      == NULL) 
                {
                c_lasmsg( FALSE, "annoteps-tmpopen",
		  "Error opening temporary file %s",
		  textfname[orient] ); 
                APS_REMTEXT;
                return E_FAIL;
                }
        }

/*  Allocate text buffer  */

    if ((textbuf = (unsigned char *) malloc( (npixhoriz > npixvert) ?
      npixhoriz : npixvert ))== NULL)
	{
	c_lasmsg( FALSE, "annoteps-malloc", "Memory allocation error" );
	APS_REMTEXT;
	return E_FAIL;
	}

/*  Open input image files  */

    imgfilecnt = 0;
    if ((subcmd == APS_INPLACE) || (subcmd == APS_INOUT))
	{
	acc = (subcmd == APS_INPLACE) ? IUPDATE : IREAD;
	imgfilecnt += input->totimage;
	if ((subcmd == APS_INOUT) && (*odtype != 0))
	    input->dtype[0] = *odtype;
	else
	    input->dtype[0] = 0;
	for (inimage = 0, band = 0; inimage < input->totimage;
	   band += input->nband[inimage++])
	    {
	    if (inimage > 0)
		{
		if (subcmd == APS_INOUT)
		    input->dtype[inimage] = input->dtype[0];
		else
		    input->dtype[inimage] = 0;
		}		
	    if (c_eopens( fdesc + inimage,
	      input->hname + inimage * CMLEN,
	      input->bands + inimage * (MAXBND + 1),
	      input->window + 4 * inimage,
	      input->window + 4 * inimage + 1,
	      input->window + 4 * inimage + 2,
	      input->window + 4 * inimage + 3,
	      input->offsets + band, &acc, input->dtype + inimage, &zero, 
	      opt ) == E_FAIL)
		{
		c_lasmsg( FALSE, "annoteps-imgopen", "Error opening %s"
		  " for input", input->hname + inimage * CMLEN );
		APS_REMTEXT;
		return E_FAIL;
		}
	    input->dlength[inimage] = dlengths[input->dtype[inimage]];
	    }
	fdesc[input->totimage] = NULL;
	}

/*  Open output image  */

    if ((subcmd == APS_INOUT) || (subcmd == APS_GENIMG))
	{
	nbandout = (subcmd == APS_INOUT) ? input->totband : 1;
	acc = IWRITE;
	if (*odtype == 0)
	    *odtype = input->dtype[0];
	odlength = dlengths[*odtype];
	if (c_eopens( fdesc + imgfilecnt++, outimg, &nbandout, &one,
	   &one, &nlo, &nso, outoffs, &acc, odtype, &zero, opt )
	  != E_SUCC)
	    {
	    c_lasmsg( FALSE, "annoteps-imgopen", "Error opening %s"
		  " for output", outimg );
	    APS_REMTEXT;
	    return E_FAIL;
	    }
	}
    else		/* For INPLACE, input == output */
	nbandout = input->totband;

/*  Allocate image buffers and set up image I/O group  */

    fdesc[imgfilecnt] = NULL;
    if ((imgbuf = (unsigned char *) c_ealloc( fdesc, &imgbufsiz ))
      == NULL)
	return E_FAIL;
    if (c_egroup( fdesc, &gdesc, imgbuf, &imgbufsiz ) != E_SUCC)
	return E_FAIL;

/*  Initialize computation of minima and maxiuma unless not wanted  */

    if (ifminmax)	/* Turned off in aps_getarg() for INPLACE  */
	if (c_inmnmx( odtype, &nbandout, minval, maxval ) != E_SUCC)
	    return E_FAIL;

/*  Start processing progress messages  */

    if (pmsgtim > 0)
	{
	opt[0] = GEN2;
	c_inittm( "ANNOTEPS", "processing", "line", "", "", &nlo, &zero,
	  &pmsgtim, opt );
	c_startm();
	}

/*  Read first input image line and initialize output buffer  */

    if (c_estep( &gdesc ) != E_SUCC)
	{
	APS_REMTEXT;
	return E_FAIL;
	}

/*  Insert text into one image line at a time  */

    dtype = *odtype;
    pixoverflag[0] = pixoverflag[1] = 0;
    bginitflag = 1;
    for (line = 1; line <= nlo; line++)
	{
	curent = line;

/*  Put pixels into output buffer unless subcommand is INPLACE  */

	if (subcmd == APS_INOUT)
	    for (band = 0; band < nbandout; band++)
		c_pxcopy( imgbuf + (odlength * input->offsets[band]),
		  imgbuf + (odlength * outoffs[band]), odtype, &nso );
	else if ((subcmd == APS_GENIMG) && (bginitflag))
	    {
	    c_pxfill( imgbuf + (odlength * outoffs[0]), odtype, &nso,
	      (unsigned char *) &outbgrnd );
	    bginitflag = 0;
	    }

/*  Insert text strings into one output line at a time  */

	for (block = 0; block < nblock; block++)
	    {
	    if ((line < text[block].slo)
	      || (line >= text[block].slo + text[block].nl))
		continue;
	    bginitflag = 1;
	    orient = (text[block].orient == APS_LTOR) ? 0 : 1;

/*  Read line of rasterized text block  */

	    fseek( textptr[orient], ((text[block].topl - 1)
	      + (line - text[block].slo)) * textwidth[orient]
	      + (text[block].lefts - 1), SEEK_SET ); 
	    if (fread( textbuf, 1, text[block].ns, textptr[orient] )
	      != text[block].ns) 
		{
		c_lasmsg( FALSE, "annoteps-readtmp",
		  "Error reading from temporary file %s",
		  textfname[orient] );
		APS_REMTEXT;
		return E_FAIL;
		}

/*  Insert rasterized text into one band at a time  */

	    for (band = iband = inimage = 0; band < nbandout; band++)
		{
		if (subcmd == APS_INPLACE)
		    {
		    dtype = input->dtype[inimage];
		    odlength = input->dlength[inimage];
		    outptr = imgbuf + odlength * input->offsets[iband];
		    }
		else
		    outptr = imgbuf + odlength * outoffs[band];
		outptr += odlength * (text[block].sso - 1);

/*  Get text pixel values from lookup table, checking for in range  */

		for (tpixel = textbuf;
		  tpixel < textbuf + text[block].ns;
		     tpixel++, outptr += odlength)
		    {
		    if ((text[block].transopt == APS_TRANS)
		      && (*tpixel == 255))
			continue;
		    textval = text[block].lookup[*tpixel];
		    if (textval > dmaxval[dtype])
			{
			textval = dmaxval[dtype];
			if (pixoverflag[0] == 0)
			    {
			    pixoverflag[0] = 1;
			    c_lasmsg( FALSE, "annoteps-overflw", "Grey"
			      " level value range exceeds maximum for"
			      " type %s -- value reduced to %d",
			      dnames[dtype], dmaxval[dtype] );
			    }
			}
		    else if (textval < dminval[dtype])
			{
			textval = dminval[dtype];
			if (pixoverflag[1] == 0)
			    {
			    pixoverflag[1] = 1;
			    c_lasmsg( FALSE, "annoteps-overflw", "Grey"
			      " level value range is less than minimum"
			      " for type%s -- value increased to %d",
			      dnames[dtype], dminval[dtype] );
			    }
			}

/*  Insert text pixel into output buffer  */

		    switch( dtype )
			{
			case EBYTE:
			  *outptr = (unsigned char) textval;
			  break;
			case EWORD:
			  *( (short *) outptr) = (short) textval;
			  break;
			case ELONG:
			  *( (long *) outptr) = (long) textval;
			  break;
			case EREAL:
			  *( (float *) outptr) = (float) textval;
			  break;
			case EDOUBLE:
			  *( (double *) outptr) = (double) textval;
			  break;
			}
		    }
		if ((subcmd == APS_INPLACE)
		  && (iband++ >= input->nband[inimage]))
		    {
		    inimage++;
		    iband = 0;
		    }
		}
	    }

/*  Update maxima and minima  */

	if (ifminmax)
	    {
	    for (band = iband = inimage = 0; band < nbandout; band++)
		{
		outptr = imgbuf + (odlength * outoffs[band]);
		c_pxmnmx( (long*) outptr, &nso, minval + band,
		  maxval +  band, &dtype );
		}
	    }

/*  Write output line(s) and read next input  */

	if (c_estep( &gdesc ) != E_SUCC)
	    return E_FAIL;
	}

/*  Stop processing messages, if used  */

    if (pmsgtim > 0)
	c_stoptm();

/*  Close all files and delete temporaries  */

    APS_REMTEXT;

/*  Copy history file and update DDR if appropriate  */

    if (subcmd == APS_INOUT)
	{
	c_hstout( vblock, outimg, (void *) input->hname,
	  &(input->totimage), "" );
	c_upddr( vblock, (void *) input->hname, outimg,
	  &(input->totimage), (void *) &(input->window[0]),
	  (void *) input->bands, input->nband, &dblone, &dblone,
	  &nbandout, &zero );
	}

/*  Enter maxima and minima into DDR  */

    
    if (subcmd != APS_INPLACE)
	c_svmnmx( outimg, &nbandout, minval, maxval, "YES" );

    return E_SUCC;
    }			/* End of Function aps_instxt */
